"""Registry for SSO backends.

Version Added:
    5.0
"""

import re
from importlib import import_module

from django.urls import include, path, re_path
from django.utils.translation import gettext_lazy as _
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         NOT_REGISTERED)

from reviewboard.registries.registry import Registry


class SSOBackendRegistry(Registry):
    """A registry for managing SSO backends.

    Version Added:
        5.0
    """

    entry_point = 'reviewboard.sso_backends'
    lookup_attrs = ['backend_id']

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered SSO backend.'
        ),
        NOT_REGISTERED: _(
            '"%(attr_value)s" is not a registered SSO backend.'
        ),
    }

    def __init__(self):
        """Initialize the registry."""
        super().__init__()
        self._url_patterns = {}

    def get_defaults(self):
        """Yield the built-in SSO backends.

        This will make sure the standard SSO backends are always present in the
        registry.

        Yields:
            reviewboard.accounts.sso.backends.base.BaseSSOBackend:
            The SSO backend instances.
        """
        builtin_backends = (
            ('saml.sso_backend', 'SAMLSSOBackend'),
        )

        for _module, _backend_cls_name in builtin_backends:
            mod = import_module('reviewboard.accounts.sso.backends.%s'
                                % _module)
            cls = getattr(mod, _backend_cls_name)
            yield cls()

    def get_siteconfig_defaults(self):
        """Return defaults for the site configuration.

        Returns:
            dict:
            The defaults to register for the site configuration.
        """
        defaults = {}

        for backend in self:
            defaults.update(backend.siteconfig_defaults)

        return defaults

    def register(self, backend):
        """Register an SSO backend.

        This also adds the URL patterns defined by the backend. If the backend
        has a
        :py:attr:`~reviewboard.accounts.sso.backends.base.SSOBackend.urls`
        attribute that is non-``None``, they will be automatically added.

        Args:
            backend (reviewboard.accounts.sso.backends.base.BaseSSOBackend):
                The backend instance.
        """
        super().register(backend)

        from reviewboard.accounts.urls import sso_dynamic_urls

        if backend.urls:
            backend_id = backend.backend_id
            backend_urls = backend.urls

            if backend.login_view_cls:
                backend_urls.append(path(
                    'login/',
                    backend.login_view_cls.as_view(sso_backend=backend),
                    name='login'))

            dynamic_urls = [
                re_path(
                    r'^(?P<backend_id>%s)/' % re.escape(backend_id),
                    include((backend_urls, 'accounts'),
                            namespace=backend_id)),
            ]
            self._url_patterns[backend_id] = dynamic_urls
            sso_dynamic_urls.add_patterns(dynamic_urls)

    def unregister(self, backend):
        """Unregister an SSO backend.

        This will remove all registered URLs that the backend has defined.

        Args:
            backend (reviewboard.accounts.sso.backends.base.BaseSSOBackend):
                The backend instance.
        """
        super().unregister(backend)

        from reviewboard.accounts.urls import sso_dynamic_urls

        try:
            dynamic_urls = self._url_patterns.pop(backend.backend_id)
            sso_dynamic_urls.remove_patterns(dynamic_urls)
        except KeyError:
            pass



"""A module exposing the search backends registry."""

from djblets.registries.importer import lazy_import_registry


#: The search backend registry.
search_backend_registry = \
    lazy_import_registry('reviewboard.search.search_backends.registry',
                         'SearchBackendRegistry')



"""Unit tests for reviewboard.attachments.mimetypes.TextMimetype.

Version Added:
    7.0.3:
    This was split off from :py:mod:`reviewboard.attachments.tests`.
"""

from __future__ import annotations

import kgb

from django.core.files.uploadedfile import SimpleUploadedFile
from django.utils.safestring import SafeText

from reviewboard.attachments.forms import UploadFileForm
from reviewboard.testing import TestCase


class TextMimetypeTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.attachments.mimetypes.TextMimetype.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    fixtures = ['test_users']

    def setUp(self):
        uploaded_file = SimpleUploadedFile(
            'test.txt',
            b'<p>This is a test</p>',
            content_type='text/plain')

        review_request = self.create_review_request(publish=True)

        form = UploadFileForm(review_request, files={
            'path': uploaded_file,
        })
        self.assertTrue(form.is_valid())

        self.file_attachment = form.create()

    def test_get_thumbnail_uncached_is_safe_text(self):
        """Testing TextMimetype.get_thumbnail string type is SafeText
        without cached thumbnail
        """
        thumbnail = self.file_attachment.thumbnail

        self.assertIsInstance(thumbnail, SafeText)

    def test_get_thumbnail_cached_is_safe_text(self):
        """Testing TextMimetype.get_thumbnail string type is SafeText with
        cached thumbnail
        """
        # Django's in-memory cache won't mangle the string types, so we can't
        # rely on just calling thumbnail twice. We have to fake it, so that
        # that we simulate the real-world behavior of getting a raw string
        # back out of a real cache.
        self.spy_on(self.file_attachment.mimetype_handler._generate_thumbnail,
                    call_fake=lambda self: '<div>My thumbnail</div>')

        thumbnail = self.file_attachment.thumbnail

        self.assertIsInstance(thumbnail, SafeText)

    def test_get_raw_thumbnail_image_url(self) -> None:
        """Testing TextMimetype.test_get_raw_thumbnail_image_url"""
        mimetype_handler = self.file_attachment.mimetype_handler
        assert mimetype_handler

        message = (
            'TextMimetype does not support generating thumbnail images.'
        )

        with self.assertRaisesMessage(NotImplementedError, message):
            mimetype_handler.get_raw_thumbnail_image_url(width=300)



"""Base communication client support for hosting services.

Version Added:
    6.0:
    This replaces the client code in the old
    :py:mod:`reviewboard.hostingsvcs.service` module.
"""

from __future__ import annotations

import json
import logging
from email.generator import _make_boundary as generate_boundary
from typing import TYPE_CHECKING
from urllib.error import URLError

from django.utils.encoding import force_bytes

from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.hostingsvcs.base.http import (HostingServiceHTTPRequest,
                                               HostingServiceHTTPResponse)
from reviewboard.scmtools.crypto_utils import decrypt_password

if TYPE_CHECKING:
    from typing import Callable

    from typelets.json import JSONValue

    from reviewboard.hostingsvcs.base.hosting_service import (
        BaseHostingService,
        HostingServiceCredentials,
    )
    from reviewboard.hostingsvcs.base.http import (FormFields,
                                                   HTTPHeaders,
                                                   UploadedFiles)
    from reviewboard.hostingsvcs.models import HostingServiceAccount


logger = logging.getLogger(__name__)


class HostingServiceClient:
    """Client for communicating with a hosting service's API.

    This implementation includes abstractions for performing HTTP operations,
    and wrappers for those to interpret responses as JSON data.

    Hosting service implementations can also include an override of this class
    to add additional checking (such as GitHub's checking of rate limit
    headers), or add higher-level API functionality.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.service` to
          :py:mod:`reviewboard.hostingsvcs.base.client`.
        * Deprecated JSON utility functions now emit deprecation warnings and
          will be removed in Review Board 8.
    """

    #: The HTTP request class to construct for HTTP requests.
    #:
    #: Subclasses can replace this if they need custom behavior when
    #: constructing or invoking the request.
    #:
    #: Version Added:
    #:     4.0
    http_request_cls: type[HostingServiceHTTPRequest] = \
        HostingServiceHTTPRequest

    #: The HTTP response class to construct HTTP responses.
    #:
    #: Subclasses can replace this if they need custom ways of formatting
    #: or interpreting response data.
    #:
    #: Version Added:
    #:     4.0
    http_response_cls: type[HostingServiceHTTPResponse] = \
        HostingServiceHTTPResponse

    #: Whether to add HTTP Basic Auth headers by default.
    #:
    #: By default, hosting services will support HTTP Basic Auth. This can be
    #: turned off if not needed.
    #:
    #: Version Added:
    #:     4.0
    use_http_basic_auth: bool = True

    #: Whether to add HTTP Digest Auth headers by default.
    #:
    #: By default, hosting services will not support HTTP Digest Auth. This
    #: can be turned on if needed.
    #:
    #: Version Added:
    #:     4.0
    use_http_digest_auth: bool = False

    ######################
    # Instance variables #
    ######################

    #: The hosting service that owns this client.
    #:
    #: Type:
    #:     The hosting service that owns this client.
    hosting_service: BaseHostingService

    def __init__(
        self,
        hosting_service: BaseHostingService,
    ) -> None:
        """Initialize the client.

        Args:
            hosting_service (reviewboard.hostingsvcs.base.hosting_service.
                             BaseHostingService):
                The hosting service that is using this client.
        """
        self.hosting_service = hosting_service

    #
    # HTTP utility methods
    #

    def http_delete(
        self,
        url: str,
        headers: (HTTPHeaders | None) = None,
        *args,
        **kwargs,
    ) -> HostingServiceHTTPResponse:
        """Perform an HTTP DELETE on the given URL.

        Version Changed:
            4.0:
            This now returns a :py:class:`~reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse` instead of a 2-tuple.

        Args:
            url (str):
                The URL to perform the request on.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The HTTP response for the request.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                There was an error performing the request, and the result is
                a raw HTTP error.
        """
        return self.http_request(url=url,
                                 headers=headers,
                                 method='DELETE',
                                 **kwargs)

    def http_get(
        self,
        url: str,
        headers: (HTTPHeaders | None) = None,
        *args,
        **kwargs,
    ) -> HostingServiceHTTPResponse:
        """Perform an HTTP GET on the given URL.

        Version Changed:
            4.0:
            This now returns a :py:class:`reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse` instead of a 2-tuple.

        Args:
            url (str):
                The URL to perform the request on.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The HTTP response for the request.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                There was an error performing the request, and the result is
                a raw HTTP error.
        """
        return self.http_request(url=url,
                                 headers=headers,
                                 method='GET',
                                 **kwargs)

    def http_head(
        self,
        url: str,
        headers: (HTTPHeaders | None) = None,
        *args,
        **kwargs,
    ) -> HostingServiceHTTPResponse:
        """Perform an HTTP HEAD on the given URL.

        Version Added:
            4.0

        Args:
            url (str):
                The URL to perform the request on.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The HTTP response for the request.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                There was an error performing the request, and the result is
                a raw HTTP error.
        """
        return self.http_request(url=url,
                                 headers=headers,
                                 method='HEAD',
                                 **kwargs)

    def http_post(
        self,
        url: str,
        body: (bytes | None) = None,
        fields: (FormFields | None) = None,
        files: (UploadedFiles | None) = None,
        content_type: (str | None) = None,
        headers: (HTTPHeaders | None) = None,
        *args,
        **kwargs,
    ) -> HostingServiceHTTPResponse:
        """Perform an HTTP POST on the given URL.

        Version Changed:
            4.0:
            This now returns a :py:class:`reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse` instead of a 2-tuple. The response can
            be treated as a 2-tuple for older code.

        Args:
            url (str):
                The URL to perform the request on.

            body (bytes, optional):
                The request body.

                If not provided, it will be generated from the ``fields`` and
                ``files`` arguments.

            fields (dict, optional):
                Form fields to use to generate the request body.

                This argument will only be used if ``body`` is ``None``.

            files (dict, optional):
                Files to use to generate the request body.

                This argument will only be used if ``body`` is ``None``.

            content_type (str, optional):
                The content type of the request.

                If provided, it will be appended as the
                :mailheader:`Content-Type` header.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The HTTP response for the request.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                There was an error performing the request, and the result is
                a raw HTTP error.
        """
        body, headers = self._build_put_post_request(body=body,
                                                     fields=fields,
                                                     files=files,
                                                     content_type=content_type,
                                                     headers=headers)

        return self.http_request(url=url,
                                 body=body,
                                 headers=headers,
                                 method='POST',
                                 **kwargs)

    def http_put(
        self,
        url: str,
        body: (bytes | None) = None,
        fields: (FormFields | None) = None,
        files: (UploadedFiles | None) = None,
        content_type: (str | None) = None,
        headers: (HTTPHeaders | None) = None,
        *args,
        **kwargs,
    ) -> HostingServiceHTTPResponse:
        """Perform an HTTP PUT on the given URL.

        Version Added:
            4.0

        Args:
            url (str):
                The URL to perform the request on.

            body (bytes, optional):
                The request body.

                If not provided, it will be generated from the ``fields`` and
                ``files`` arguments.

            fields (dict, optional):
                Form fields to use to generate the request body.

                This argument will only be used if ``body`` is ``None``.

            files (dict, optional):
                Files to use to generate the request body.

                This argument will only be used if ``body`` is ``None``.

            content_type (str, optional):
                The content type of the request.

                If provided, it will be appended as the
                :mailheader:`Content-Type` header.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The HTTP response for the request.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                There was an error performing the request, and the result is
                a raw HTTP error.
        """
        body, headers = self._build_put_post_request(body=body,
                                                     fields=fields,
                                                     files=files,
                                                     content_type=content_type,
                                                     headers=headers)

        return self.http_request(url=url,
                                 body=body,
                                 headers=headers,
                                 method='PUT',
                                 **kwargs)

    def http_request(
        self,
        url: str,
        body: (bytes | None) = None,
        headers: (HTTPHeaders | None) = None,
        method: str = 'GET',
        **kwargs,
    ) -> HostingServiceHTTPResponse:
        """Perform an HTTP request, processing and handling results.

        This constructs an HTTP request based on the specified criteria,
        returning the resulting data and headers or raising a suitable error.

        In most cases, callers will use one of the wrappers, like
        :py:meth:`http_get` or :py:meth:`http_post`. Calling this directly is
        useful if working with non-standard HTTP methods.

        Subclasses can control the behavior of HTTP requests through several
        related methods:

        * :py:meth:`get_http_credentials`
          - Return credentials for use in the HTTP request.

        * :py:meth:`build_http_request`
          - Build the :py:class:`reviewboard.hostingsvcs.base.http.
          HostingServiceHTTPRequest` object.

        * :py:meth:`open_http_request`
          - Performs the actual HTTP request.

        * :py:meth:`process_http_response`
          - Performs post-processing on a response from the service, or raises
          an error.

        * :py:meth:`process_http_error`
          - Processes a raised exception, handling it in some form or
          converting it into another error.

        See those methods for more information.

        Version Changed:
            4.0:
            This now returns a :py:class:`reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse` instead of a 2-tuple.

        Args:
            url (str):
                The URL to open.

            body (bytes, optional):
                The request body.

            headers (dict, optional):
                Headers to include in the request.

            method (str, optional):
                The HTTP method to use to perform the request.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`build_http_request`.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The HTTP response for the request.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                There was an error performing the request, and the result is
                a raw HTTP error.
        """
        credentials = self.get_http_credentials(
            account=self.hosting_service.account,
            **kwargs)

        request = self.build_http_request(url=url,
                                          body=body,
                                          headers=headers,
                                          method=method,
                                          credentials=credentials,
                                          **kwargs)

        try:
            return self.process_http_response(self.open_http_request(request))
        except URLError as e:
            # This will either raise, or it will return and we'll raise.
            self.process_http_error(request, e)

            raise

    def get_http_credentials(
        self,
        account: HostingServiceAccount,
        username: (str | None) = None,
        password: (str | None) = None,
        **kwargs,
    ) -> HostingServiceCredentials:
        """Return credentials used to authenticate with the service.

        Subclasses can override this to return credentials based on the
        account or the values passed in when performing the HTTP request.
        The resulting dictionary contains keys that will be processed in
        :py:meth:`build_http_request`.

        There are a few supported keys that subclasses will generally want
        to return:

        Keys:
            username (str):
                The username, typically for use in HTTP Basic Auth or HTTP
                Digest Auth.

            password (str):
                The accompanying password.

            headers (dict):
                A dictionary of authentication headers to add to the request.

        By default, this will return a ``username`` and ``password`` based on
        the request (if those values are provided by the caller).

        Args:
            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The stored authentication data for the service.

            username (str, optional):
                An explicit username passed by the caller.

                This will override the data stored in the account, if both a
                username and password are provided.

            password (str, optional):
                An explicit password passed by the caller.

                This will override the data stored in the account, if both a
                username and password are provided.

            **kwargs (dict, unused):
                Additional keyword arguments passed in when making the HTTP
                request.

        Returns:
            dict:
            A dictionary of credentials for the request.
        """
        if (username is None and
            password is None and
            'password' in account.data):
            username = account.username
            password = decrypt_password(account.data['password'])

        if username is not None and password is not None:
            return {
                'username': username,
                'password': password,
            }

        return {}

    def open_http_request(
        self,
        request: HostingServiceHTTPRequest,
    ) -> HostingServiceHTTPResponse:
        """Perform a raw HTTP request and return the result.

        This is not meant to be called directly. Please use one of the
        following methods instead:

        * :py:meth:`http_delete`
        * :py:meth:`http_get`
        * :py:meth:`http_head`
        * :py:meth:`http_post`
        * :py:meth:`http_put`
        * :py:meth:`http_request`

        Args:
            request (reviewboard.hostingsvcs.base.http.
                     HostingServiceHTTPRequest):
                The HTTP request to open.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The successful response information from the server.

        Raises:
            urllib.error.URLError:
                There was an error performing a request on the URL.
        """
        return request.open()

    def build_http_request(
        self,
        credentials: HostingServiceCredentials,
        **kwargs,
    ) -> HostingServiceHTTPRequest:
        """Build a request object for an HTTP request.

        This constructs a :py:class:`HostingServiceHTTPRequest` containing
        the information needed to perform the HTTP request by passing the
        provided keyword arguments to the the constructor.

        If ``username`` and ``password`` are provided in ``credentials``, this
        will also add a HTTP Basic Auth header (if
        :py:attr:`use_http_basic_auth` is set) and HTTP Digest Auth Header (if
        :py:attr:`use_http_digest_auth` is set).

        Subclasses can override this to change any behavior as needed. For
        instance, adding other headers or authentication schemes.

        Args:
            credentials (dict):
                The credentials used for the request.

            **kwargs (dict, unused):
                Keyword arguments for the :py:class:`reviewboard.hostingsvcs.
                base.http.HostingServiceHTTPRequest` instance.

        Returns:
            HostingServiceHTTPRequest:
            The resulting request object for use in the HTTP request.
        """
        request = self.http_request_cls(hosting_service=self.hosting_service,
                                        **kwargs)

        if credentials:
            username = credentials.get('username')
            password = credentials.get('password')

            if username is not None and password is not None:
                if self.use_http_basic_auth:
                    request.add_basic_auth(username, password)

                if self.use_http_digest_auth:
                    request.add_digest_auth(username, password)

            auth_headers = credentials.get('headers') or {}

            for header, value in auth_headers.items():
                request.add_header(header, value)

        return request

    def process_http_response(
        self,
        response: HostingServiceHTTPResponse,
    ) -> HostingServiceHTTPResponse:
        """Process an HTTP response and return a result.

        This can be used by subclasses to modify a response before it gets back
        to the caller. It can also raise a :py:class:`urllib.error.URLError`
        (which will get processed by :py:meth:`process_http_error`), or a
        :py:class:`~reviewboard.hostingsvcs.errors.HostingServiceError`.

        By default, the response is returned as-is.

        Args:
            response (reviewboard.hostingsvcs.base.http.
                      HostingServiceHTTPResponse):
                The response to process.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The resulting response.
        """
        return response

    def process_http_error(
        self,
        request: HostingServiceHTTPRequest,
        e: URLError,
    ) -> None:
        """Process an HTTP error, possibly raising a result.

        By default, this doesn't do anything.

        Subclasses can override this to provide more specific errors as needed
        by the hosting service implementation. They should always call the
        parent method as well.

        If there's no specific exception, this should just return, allowing
        the original exception to be raised.

        Args:
            request (reviewboard.hostingsvcs.base.http.
                     HostingServiceHTTPRequest):
                The request that resulted in an error.

            e (urllib.error.URLError):
                The error to process.
        """
        pass

    #
    # JSON utility methods
    #

    def json_delete(
        self,
        *args,
        **kwargs,
    ) -> tuple[JSONValue, HTTPHeaders]:
        """Perform an HTTP DELETE and interpret the results as JSON.

        Deprecated:
            4.0:
            Use :py:meth:`http_delete` instead, and access the
            :py:attr:`~reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse.json` attribute on the response for the
            JSON payload.

        Args:
            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_delete`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_delete`.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (object):
                    The JSON data (in the appropriate type).

                1 (dict):
                    The HTTP response headers.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                When there is an error communicating with the URL.
        """
        RemovedInReviewBoard90Warning.warn(
            '%(class_name)s.json_delete is deprecated. Please use '
            '%(class_name)s.http_delete instead. This will be removed in '
            'Review Board 9.'
            % {
                'class_name': type(self).__name__,
            })

        return self._do_json_method(self.http_delete, *args, **kwargs)

    def json_get(
        self,
        *args,
        **kwargs,
    ) -> tuple[JSONValue, HTTPHeaders]:
        """Perform an HTTP GET and interpret the results as JSON.

        Deprecated:
            4.0:
            Use :py:meth:`http_get` instead, and access the
            :py:attr:`~reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse.json` attribute on the response for
            the JSON payload.

        Args:
            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_get`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_get`.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (object):
                    The JSON data (in the appropriate type).

                1 (dict):
                    The HTTP response headers.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                When there is an error communicating with the URL.
        """
        RemovedInReviewBoard90Warning.warn(
            '%(class_name)s.json_get is deprecated. Please use '
            '%(class_name)s.http_get instead. This will be removed in '
            'Review Board 9.'
            % {
                'class_name': type(self).__name__,
            })

        return self._do_json_method(self.http_get, *args, **kwargs)

    def json_post(
        self,
        *args,
        **kwargs,
    ) -> tuple[JSONValue, HTTPHeaders]:
        """Perform an HTTP POST and interpret the results as JSON.

        Deprecated:
            4.0:
            Use :py:meth:`http_post` instead, and access the
            :py:attr:`~reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse.json` attribute on the response for
            the JSON payload.

        Args:
            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_post`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_post`.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (object):
                    The JSON data (in the appropriate type).

                1 (dict):
                    The HTTP response headers.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                When there is an error communicating with the URL.
        """
        RemovedInReviewBoard90Warning.warn(
            '%(class_name)s.json_post is deprecated. Please use '
            '%(class_name)s.http_post instead. This will be removed in '
            'Review Board 9.'
            % {
                'class_name': type(self).__name__,
            })

        return self._do_json_method(self.http_post, *args, **kwargs)

    def _do_json_method(
        self,
        method: Callable[
            ...,
            tuple[bytes, HTTPHeaders] | HostingServiceHTTPResponse
        ],
        *args,
        **kwargs,
    ) -> tuple[JSONValue, HTTPHeaders]:
        """Parse the result of an HTTP operation as JSON.

        Deprecated:
            4.0:
            Use :py:meth:`http_post` instead, and access the
            :py:attr:`~reviewboard.hostingsvcs.base.http.
            HostingServiceHTTPResponse.json` attribute on the response for
            the JSON payload.

        Args:
            method (callable):
                The callable to use to execute the request.

            *args (tuple):
                Positional arguments to pass to ``method``.

            **kwargs (dict):
                Keyword arguments to pass to ``method``.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (object):
                    The JSON data (in the appropriate type).

                1 (dict):
                    The HTTP response headers.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error performing the request, and the error has
                been translated to a more specific hosting service error.

            urllib.error.URLError:
                When there is an error communicating with the URL.
        """
        response = method(*args, **kwargs)

        if isinstance(response, HostingServiceHTTPResponse):
            return response.json, response.headers
        elif isinstance(response, tuple):
            data, headers = response

            if data:
                data = json.loads(data.decode('utf-8'))

            return data, headers
        else:
            raise NotImplementedError(f'Unsupported response from {method!r}')

    def _build_put_post_request(
        self,
        body: (bytes | None) = None,
        fields: (FormFields | None) = None,
        files: (UploadedFiles | None) = None,
        content_type: (str | None) = None,
        headers: (HTTPHeaders | None) = None,
    ) -> tuple[bytes, HTTPHeaders]:
        """Build a request body and headers for a HTTP PUT or POST.

        Args:
            body (bytes, optional):
                The request body content.

            fields (dict, optional):
                The form fields used to construct a request body.

                This is ignored if ``body`` is set.

            files (dict, optional):
                The uploaded files used to construct a request body.

                This is ignored if ``body`` is set.

            content_type (str, optional):
                The value used for the :mailheader:`Content-Type` header.

            headers (dict, optional):
                Additional headers to set in the request.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (bytes):
                    The body to post.

                1 (dict):
                    The HTTP response headers.
        """
        if headers:
            headers = headers.copy()
        else:
            headers = {}

        if body is None:
            if fields is not None or files is not None:
                body, content_type = self.build_form_data(fields, files)
            else:
                body = b''
        else:
            body = force_bytes(body)

        if content_type:
            headers['Content-Type'] = content_type

        headers['Content-Length'] = str(len(body))

        return body, headers

    #
    # Internal utilities
    #

    @staticmethod
    def build_form_data(
        fields: FormFields | None,
        files: (UploadedFiles | None) = None,
    ) -> tuple[bytes, str]:
        """Encode data for use in an HTTP POST.

        Args:
            fields (dict):
                A mapping of field names to values.

            files (dict, optional):
                A mapping of field names to files (:py:class:`dict`).

        Returns:
            A 2-tuple of:

            Tuple:
                0 (bytes):
                    The body to post.

                1 (dict):
                    The HTTP response headers.
        """
        boundary = HostingServiceClient._make_form_data_boundary()
        enc_boundary = boundary.encode('utf-8')
        content_parts = []

        if fields:
            for key, value in sorted(fields.items(),
                                     key=lambda pair: pair[0]):
                if isinstance(key, str):
                    key = key.encode('utf-8')

                if isinstance(value, str):
                    value = value.encode('utf-8')

                content_parts.append(
                    b'--%(boundary)s\r\n'
                    b'Content-Disposition: form-data; name="%(key)s"\r\n'
                    b'\r\n'
                    b'%(value)s\r\n'
                    % {
                        b'boundary': enc_boundary,
                        b'key': key,
                        b'value': value,
                    }
                )

        if files:
            for key, data in sorted(files.items(),
                                    key=lambda pair: pair[0]):
                filename = data['filename']
                content = data['content']

                if isinstance(key, str):
                    key = key.encode('utf-8')

                if isinstance(filename, str):
                    filename = filename.encode('utf-8')

                if isinstance(content, str):
                    content = content.encode('utf-8')

                content_parts.append(
                    b'--%(boundary)s\r\n'
                    b'Content-Disposition: form-data; name="%(key)s";'
                    b' filename="%(filename)s"\r\n'
                    b'\r\n'
                    b'%(value)s\r\n'
                    % {
                        b'boundary': enc_boundary,
                        b'key': key,
                        b'filename': filename,
                        b'value': content,
                    }
                )

        content_parts.append(b'--%s--' % enc_boundary)

        content = b''.join(content_parts)
        content_type = f'multipart/form-data; boundary={boundary}'

        return content, content_type

    @staticmethod
    def _make_form_data_boundary() -> str:
        """Return a unique boundary to use for HTTP form data.

        This primary exists for the purpose of spying in unit tests.

        Returns:
            str:
            The boundary for use in the form data.
        """
        return generate_boundary()



"""Actions for the reviews app."""

from __future__ import annotations

from typing import Iterable, List, Optional, TYPE_CHECKING, Union

from django.http import HttpRequest
from django.template import Context
from django.utils.translation import gettext_lazy as _
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.actions import (ActionPlacement,
                                 AttachmentPoint,
                                 BaseAction,
                                 BaseMenuAction,
                                 actions_registry)
from reviewboard.actions.renderers import (BaseActionRenderer,
                                           DetailedMenuActionGroupRenderer,
                                           MenuActionGroupRenderer)
from reviewboard.admin.read_only import is_site_read_only_for
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.reviews.features import (general_comments_feature,
                                          unified_banner_feature)
from reviewboard.reviews.models import ReviewRequest
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.urls import all_review_request_url_names

if TYPE_CHECKING:
    from typelets.django.strings import StrOrPromise


class CloseMenuAction(BaseMenuAction):
    """A menu for closing the review request.

    Version Added:
        6.0
    """

    action_id = 'close-menu'
    label = _('Close')
    apply_to = all_review_request_url_names

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        review_request = context.get('review_request')
        perms = context.get('perms')
        user = request.user

        return (super().should_render(context=context) and
                review_request is not None and
                review_request.status == ReviewRequest.PENDING_REVIEW and
                not is_site_read_only_for(user) and
                (request.user.pk == review_request.submitter_id or
                (bool(perms) and
                 perms['reviews']['can_change_status'] and
                 review_request.public)))


class CloseCompletedAction(BaseAction):
    """The action to close a review request as completed.

    Version Added:
        6.0
    """

    action_id = 'close-completed'
    label = _('Completed')
    apply_to = all_review_request_url_names
    js_model_class = 'RB.CloseCompletedAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=CloseMenuAction.action_id),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        review_request = context.get('review_request')

        return (super().should_render(context=context) and
                review_request is not None and
                review_request.public and
                not is_site_read_only_for(context['request'].user))


class CloseDiscardedAction(BaseAction):
    """The action to close a review request as discarded.

    Version Added:
        6.0
    """

    action_id = 'close-discarded'
    label = _('Discarded')
    apply_to = all_review_request_url_names
    js_model_class = 'RB.CloseDiscardedAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=CloseMenuAction.action_id),
    ]


class DeleteAction(BaseAction):
    """The action to permanently delete a review request.

    Version Added:
        6.0
    """

    action_id = 'delete-review-request'
    label = _('Delete Permanently')
    apply_to = all_review_request_url_names
    js_model_class = 'RB.DeleteAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=CloseMenuAction.action_id),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        perms = context.get('perms')

        return (super().should_render(context=context) and
                bool(perms) and
                perms['reviews']['delete_reviewrequest'] and
                not is_site_read_only_for(context['request'].user))


class DownloadDiffAction(BaseAction):
    """The action to download a diff.

    Version Added:
        6.0
    """

    action_id = 'download-diff'
    label = _('Download Diff')
    apply_to = all_review_request_url_names

    def get_url(
        self,
        *,
        context: Context,
    ) -> str:
        """Return this action's URL.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            str:
            The URL to invoke if this action is clicked.
        """
        # We want to use a relative URL in the diff viewer as we will not be
        # re-rendering the page when switching between revisions.
        from reviewboard.urls import diffviewer_url_names

        request: HttpRequest = context['request']
        match = request.resolver_match

        if match and match.url_name in diffviewer_url_names:
            return 'raw/'

        return local_site_reverse(
            'raw-diff',
            request,
            kwargs={
                'review_request_id': context['review_request'].display_id,
            })

    def get_visible(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether the action should start visible or not.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should start visible. ``False``, otherwise.
        """
        from reviewboard.urls import diffviewer_url_names

        request: HttpRequest = context['request']
        match = request.resolver_match

        if match and match.url_name in diffviewer_url_names:
            return match.url_name != 'view-interdiff'

        return super().get_visible(context=context)

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        from reviewboard.urls import diffviewer_url_names

        request: HttpRequest = context['request']
        match = request.resolver_match

        # If we're on a diff viewer page, then this should be initially
        # rendered, but might be hidden.
        if match and match.url_name in diffviewer_url_names:
            return True

        review_request = context.get('review_request')

        return (super().should_render(context=context) and
                review_request is not None and
                review_request.has_diffsets)


class ReviewMenuActionRenderer(DetailedMenuActionGroupRenderer):
    """Action renderer for the Review menu.

    This renders as a detailed menu, but using a custom JavaScript view for
    managing the menu.

    Version Added:
        8.0
    """

    js_view_class = 'RB.ReviewMenuActionView'


class ReviewMenuAction(BaseMenuAction):
    """The "Review" menu on the unified banner.

    Version Added:
        6.0
    """

    action_id = 'review-menu'
    apply_to = all_review_request_url_names
    default_renderer_cls = ReviewMenuActionRenderer
    icon_class = 'rb-icon rb-icon-compose-review'
    label = _('Review')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.UNIFIED_BANNER),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This menu only renders when the user is logged in and the unified
        banner feature is enabled.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user

        return (super().should_render(context=context) and
                user.is_authenticated and
                not is_site_read_only_for(user) and
                unified_banner_feature.is_enabled(request=request))


class CreateReviewAction(BaseAction):
    """Action to create a new, blank review.

    Version Added:
        6.0
    """

    action_id = 'create-review'
    apply_to = all_review_request_url_names
    label = _('Create Review')
    verbose_label = _('Create a new review')
    description = [
        _('Your review will start off blank, but you can add text and '
          'general comments to it.'),
        _('Adding comments to code or file attachments will automatically '
          'create a new review for you.'),
    ]
    icon_class = 'rb-icon rb-icon-create-review'
    js_model_class = 'RB.CreateReviewAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.UNIFIED_BANNER,
                        parent_id='review-menu'),
        ActionPlacement(attachment=AttachmentPoint.QUICK_ACCESS),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This menu only renders when the user is logged in and the unified
        banner feature is enabled.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user

        return (super().should_render(context=context) and
                user.is_authenticated and
                not is_site_read_only_for(user) and
                unified_banner_feature.is_enabled(request=request))


class EditReviewAction(BaseAction):
    """Action to edit an existing review.

    Version Added:
        6.0
    """

    action_id = 'edit-review'
    apply_to = all_review_request_url_names
    label = _('Edit Review')
    verbose_label = _('Edit your review')
    description = [
        _('Edit your comments and publish your review.'),
    ]
    icon_class = 'rb-icon rb-icon-compose-review'
    js_model_class = 'RB.EditReviewAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.UNIFIED_BANNER,
                        parent_id='review-menu'),
        ActionPlacement(attachment=AttachmentPoint.QUICK_ACCESS),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This menu only renders when the user is logged in and the unified
        banner feature is enabled.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user

        return (super().should_render(context=context) and
                user.is_authenticated and
                not is_site_read_only_for(user) and
                unified_banner_feature.is_enabled(request=request))


class AddGeneralCommentAction(BaseAction):
    """Action to add a general comment.

    Version Added:
        6.0
    """

    action_id = 'add-general-comment'
    apply_to = all_review_request_url_names
    label = _('Add General Comment')
    verbose_label = _('Add a general comment')
    description = [
        _('Add a new general comment about the change, not attached to '
          'any code or file attachments.'),
    ]
    icon_class = 'rb-icon rb-icon-edit'
    js_model_class = 'RB.AddGeneralCommentAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.UNIFIED_BANNER,
                        parent_id='review-menu'),
        ActionPlacement(attachment=AttachmentPoint.QUICK_ACCESS),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        Version Added:
            8.0

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user

        return (super().should_render(context=context) and
                user.is_authenticated and
                not is_site_read_only_for(user))


class ShipItAction(BaseAction):
    """Action to mark a review request as "Ship It".

    Version Added:
        6.0
    """

    action_id = 'ship-it'
    apply_to = all_review_request_url_names
    label = _('Ship It!')
    verbose_label = _('Ship it!')
    description = [
        _("You're happy with what you're seeing, and would like to "
          'approve it.'),
        _('If you want to leave a comment with this, choose "Create '
          'a new review" above.'),
    ]
    icon_class = 'rb-icon rb-icon-shipit'
    js_model_class = 'RB.ShipItAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.UNIFIED_BANNER,
                        parent_id='review-menu'),
        ActionPlacement(attachment=AttachmentPoint.QUICK_ACCESS),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        Version Added:
            8.0

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user
        review_request = context.get('review_request')
        siteconfig = SiteConfiguration.objects.get_current()

        return (
            super().should_render(context=context) and
            user.is_authenticated and
            not is_site_read_only_for(user) and
            review_request is not None and
            (user.pk != review_request.submitter_id or
             bool(siteconfig.get('reviews_allow_self_shipit')))
        )


class LegacyAddGeneralCommentAction(BaseAction):
    """The action for adding a general comment.

    Deprecated:
        8.0:
        This will be removed in Review Board 9.

    Version Added:
        6.0
    """

    action_id = 'legacy-add-general-comment'
    label = _('Add General Comment')
    apply_to = all_review_request_url_names

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user

        return (super().should_render(context=context) and
                user.is_authenticated and
                not is_site_read_only_for(user) and
                general_comments_feature.is_enabled(request=request) and
                not unified_banner_feature.is_enabled(request=request))


class LegacyEditReviewAction(BaseAction):
    """The old-style "Edit Review" action.

    This exists within the review request actions area, and will be supplanted
    by the new action in the Review menu in the unified banner.

    Deprecated:
        8.0:
        This will be removed in Review Board 9.

    Version Added:
        6.0
    """

    action_id = 'legacy-edit-review'
    label = _('Review')
    apply_to = all_review_request_url_names

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user

        return (super().should_render(context=context) and
                user.is_authenticated and
                not is_site_read_only_for(user) and
                not unified_banner_feature.is_enabled(request=request))


class LegacyShipItAction(BaseAction):
    """The old-style "Ship It" action.

    This exists within the review request actions area, and will be supplanted
    by the new action in the Review menu in the unified banner.

    Deprecated:
        8.0:
        This will be removed in Review Board 9.

    Version Added:
        6.0
    """

    action_id = 'legacy-ship-it'
    label = _('Ship It!')
    apply_to = all_review_request_url_names

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        user = request.user

        return (super().should_render(context=context) and
                user.is_authenticated and
                not is_site_read_only_for(user) and
                not unified_banner_feature.is_enabled(request=request))


class UpdateMenuAction(BaseMenuAction):
    """A menu for updating the review request.

    Version Added:
        6.0
    """

    action_id = 'update-menu'
    label = _('Update')
    apply_to = all_review_request_url_names

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        review_request = context.get('review_request')
        perms = context.get('perms')
        user = request.user

        return (super().should_render(context=context) and
                review_request is not None and
                review_request.status == ReviewRequest.PENDING_REVIEW and
                not is_site_read_only_for(user) and
                (user.pk == review_request.submitter_id or
                (bool(perms) and perms['reviews']['can_edit_reviewrequest'])))


class UploadDiffAction(BaseAction):
    """The action to update or upload a diff.

    Version Added:
        6.0
    """

    action_id = 'upload-diff'
    apply_to = all_review_request_url_names
    js_model_class = 'RB.UpdateDiffAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=UpdateMenuAction.action_id),
    ]

    def get_label(
        self,
        *,
        context: Context,
    ) -> StrOrPromise:
        """Return the label for the action.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            str:
            The label to use for the action.
        """
        review_request = context['review_request']
        draft = review_request.get_draft(context['request'].user)

        if (draft and draft.diffset) or review_request.has_diffsets:
            return _('Update Diff')
        else:
            return _('Upload Diff')

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        review_request = context.get('review_request')
        perms = context.get('perms')
        user = request.user

        return (super().should_render(context=context) and
                review_request is not None and
                review_request.repository_id is not None and
                not is_site_read_only_for(user) and
                (user.pk == review_request.submitter_id or
                 (bool(perms) and
                  perms['reviews']['can_edit_reviewrequest'])))


class UploadFileAction(BaseAction):
    """The action to upload a new file attachment.

    Version Added:
        6.0
    """

    action_id = 'upload-file'
    label = _('Add File')
    apply_to = all_review_request_url_names
    js_model_class = 'RB.AddFileAction'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=UpdateMenuAction.action_id),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        review_request = context.get('review_request')
        perms = context.get('perms')
        user = request.user

        return (super().should_render(context=context) and
                review_request is not None and
                review_request.status == ReviewRequest.PENDING_REVIEW and
                not is_site_read_only_for(user) and
                (user.pk == review_request.submitter_id or
                 (bool(perms) and
                  perms['reviews']['can_edit_reviewrequest'])))


class StarActionRenderer(BaseActionRenderer):
    """Action renderer for the starred action.

    This provides a custom template used to render the action as a star,
    which can be pressed to star or unstar the review request.

    Version Added:
        8.0
    """

    template_name = 'reviews/star_action.html'


class StarAction(BaseAction):
    """The action to star a review request.

    Version Added:
        6.0
    """

    action_id = 'star-review-request'
    apply_to = all_review_request_url_names
    default_renderer_cls = StarActionRenderer
    label = ''

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST_LEFT),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        review_request = context.get('review_request')
        user = request.user

        return (user.is_authenticated and
                review_request is not None and
                review_request.public and
                not is_site_read_only_for(user) and
                super().should_render(context=context))


class ArchiveMenuActionRenderer(MenuActionGroupRenderer):
    """Action renderer for the Archive menu.

    This renders as a menu, but using a custom JavaScript view for managing the
    menu.

    Version Added:
        8.0
    """

    js_view_class = 'RB.ArchiveMenuActionView'


class ArchiveMenuAction(BaseMenuAction):
    """A menu for managing the visibility state of the review request.

    Version Added:
        6.0
    """

    action_id = 'archive-menu'
    default_renderer_cls = ArchiveMenuActionRenderer
    apply_to = all_review_request_url_names

    icon_class = 'rb-icon rb-icon-archive-off'
    verbose_label = _('Archive or Mute')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST_LEFT),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`hidden` in that hidden actions still render
        but are hidden by CSS, whereas if this returns ``False`` the action
        will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        review_request = context.get('review_request')
        user = request.user

        return (user.is_authenticated and
                review_request is not None and
                review_request.public and
                not is_site_read_only_for(user) and
                super().should_render(context=context))


class ArchiveAction(BaseAction):
    """An action for archiving the review request.

    Version Added:
        6.0
    """

    action_id = 'archive'
    apply_to = all_review_request_url_names
    js_model_class = 'RB.ArchiveAction'
    label = _('Toggle Archived')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST_LEFT,
                        parent_id=ArchiveMenuAction.action_id),
    ]


class MuteAction(BaseAction):
    """An action for muting the review request.

    Version Added:
        6.0
    """

    action_id = 'mute'
    apply_to = all_review_request_url_names
    js_model_class = 'RB.MuteAction'
    label = _('Toggle Muted')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST_LEFT,
                        parent_id=ArchiveMenuAction.action_id),
    ]


class BaseReviewRequestAction(BaseAction):
    """A base class for an action that can be applied to a review request.

    Creating an action requires subclassing :py:class:`BaseReviewRequestAction`
    and overriding any fields/methods as desired. Different instances of the
    same subclass can also override the class fields with their own instance
    fields.

    Example:
        .. code-block:: python

           class UsedOnceAction(BaseReviewRequestAction):
               action_id = 'once'
               label = 'This is used once.'

           class UsedMultipleAction(BaseReviewRequestAction):
               def __init__(self, action_id, label):
                   super().__init__()

                   self.action_id = 'repeat-' + action_id
                   self.label = 'This is used multiple times,'

    Note:
        Since the same action will be rendered for multiple different users in
        a multithreaded environment, the action state should not be modified
        after initialization. If we want different action attributes at
        runtime, then we can override one of the getter methods (such as
        :py:meth:`get_label`), which by default will simply return the original
        attribute from initialization.

    Deprecated:
        6.0:
        New code should be written using
        :py:class:`reviewboard.actions.base.BaseAction`. This class will be
        removed in 8.0.
    """

    apply_to = all_review_request_url_names

    _ignore_action_deprecations = True

    def __init__(self) -> None:
        """Initialize this action.

        By default, actions are top-level and have no children.
        """
        RemovedInReviewBoard90Warning.warn(
            'BaseReviewRequestAction is deprecated and will be removed in '
            'Review Board 9. Please update your code to use '
            'reviewboard.actions.base.BaseAction')

        super().__init__()

        self.placements = [
            ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST),
        ]

    @property
    def max_depth(self) -> int:
        """Lazily compute the max depth of any action contained by this action.

        Top-level actions have a depth of zero, and child actions have a depth
        that is one more than their parent action's depth.

        Algorithmically, the notion of max depth is equivalent to the notion of
        height in the context of trees (from graph theory). We decided to use
        this term instead so as not to confuse it with the dimensional height
        of a UI element.

        Returns:
            int:
            The max depth of any action contained by this action.
        """
        return 0

    def reset_max_depth(self) -> None:
        """Reset the max_depth of this action and all its ancestors to zero."""
        # The max depth is now calculated on the fly, so this is a no-op.
        pass

    def get_dom_element_id(self) -> str:
        """Return the ID used for the DOM element for this action.

        Returns:
            str:
            The ID used for the element.
        """
        return f'action-{self.action_id}'

    def get_extra_context(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> dict:
        """Return extra template context for the action.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            Extra context to use when rendering the action's template.
        """
        data = super().get_extra_context(request=request, context=context)

        if self.template_name != 'actions/action.html':
            data['action'] = self.copy_to_dict(context)

        return data

    def copy_to_dict(
        self,
        context: Context,
    ) -> dict:
        """Copy this action instance to a dictionary.

        This is a legacy implementation left to maintain compatibility with
        custom templates.
        """
        return {
            'action_id': self.action_id,
            'hidden': not self.get_visible(context=context),
            'label': self.get_label(context=context),
            'url': self.get_url(context=context),
        }

    def register(
        self,
        parent: Optional[BaseReviewRequestMenuAction] = None,
    ) -> None:
        """Register this review request action instance.

        Note:
           Newly registered top-level actions are appended to the left of the
           other previously registered top-level actions. So if we intend to
           register a collection of top-level actions in a certain order, then
           we likely want to iterate through the actions in reverse.

        Args:
            parent (BaseReviewRequestMenuAction, optional):
                The parent action instance of this action instance.

        Raises:
            KeyError:
                A second registration is attempted (action IDs must be unique
                across all types of actions and menu actions, at any depth).

            DepthLimitExceededError:
                The maximum depth limit is exceeded.
        """
        if parent is not None:
            assert self.placements

            self.placements[0].parent_id = parent.action_id

        actions_registry.register(self)

    def unregister(self) -> None:
        """Unregister this review request action instance.

        Note:
           This method can mutate its parent's child actions. So if we are
           iteratively unregistering a parent's child actions, then we should
           consider first making a clone of the list of children.

        Raises:
            KeyError:
                An unregistration is attempted before it's registered.
        """
        actions_registry.unregister(self)


class BaseReviewRequestMenuAction(BaseMenuAction):
    """A base class for an action with a dropdown menu.

    Deprecated:
        6.0:
        New code should be written using
        :py:class:`reviewboard.actions.base.BaseMenuAction`. This class will be
        removed in 8.0.
    """

    apply_to = all_review_request_url_names

    _ignore_action_deprecations = True

    def __init__(
        self,
        child_actions: Optional[List[BaseReviewRequestAction]] = None,
    ) -> None:
        """Initialize this menu action.

        Args:
            child_actions (list of BaseReviewRequestAction, optional):
                The list of child actions to be contained by this menu action.

        Raises:
            KeyError:
                A second registration is attempted (action IDs must be unique
                across all types of actions and menu actions, at any depth).

            DepthLimitExceededError:
                The maximum depth limit is exceeded.
        """
        super().__init__()

        self.placements = [
            ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST),
        ]

        self._children = child_actions or []

    def get_dom_element_id(self) -> str:
        """Return the ID used for the DOM element for this action.

        Returns:
            str:
            The ID used for the element.
        """
        return f'action-{self.action_id}'

    def copy_to_dict(
        self,
        context: Context,
    ) -> dict:
        """Copy this menu action instance to a dictionary.

        This is a legacy implementation left to maintain compatibility with
        custom templates.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            dict:
            The corresponding dictionary.
        """
        return {
            'action_id': self.action_id,
            'child_actions': self.child_actions,
            'hidden': not self.get_visible(context=context),
            'label': self.get_label(context=context),
            'url': self.get_url(context=context),
        }

    def get_extra_context(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> dict:
        """Return extra template context for the action.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            Extra context to use when rendering the action's template.
        """
        data = super().get_extra_context(request=request, context=context)

        if self.template_name != 'actions/menu_action.html':
            data['menu_action'] = self.copy_to_dict(context)

        return data

    @property
    def max_depth(self) -> int:
        """Lazily compute the max depth of any action contained by this action.

        Returns:
            int: The max depth of any action contained by this action.
        """
        if self.child_actions:
            return max(child_action.max_depth
                       for child_action in self.child_actions)
        else:
            return self.depth

    def register(
        self,
        parent: Optional[BaseReviewRequestMenuAction] = None,
    ) -> None:
        """Register this review request action instance.

        Note:
           Newly registered top-level actions are appended to the left of the
           other previously registered top-level actions. So if we intend to
           register a collection of top-level actions in a certain order, then
           we likely want to iterate through the actions in reverse.

        Args:
            parent (BaseReviewRequestMenuAction, optional):
                The parent action instance of this action instance.

        Raises:
            KeyError:
                A second registration is attempted (action IDs must be unique
                across all types of actions and menu actions, at any depth).

            DepthLimitExceededError:
                The maximum depth limit is exceeded.
        """
        action_id = self.action_id

        if parent is not None:
            assert self.placements
            self.placements[0].parent_id = parent.action_id

        actions_registry.register(self)

        for child in self._children:
            assert child.placements

            child.placements[0].parent_id = action_id
            child.register()

    def unregister(self) -> None:
        """Unregister this review request action instance.

        This menu action recursively unregisters its child action instances.

        Raises:
            KeyError: An unregistration is attempted before it's registered.
        """
        for child in self._children:
            child.unregister()

        actions_registry.unregister(self)


def register_actions(
    actions: List[Union[BaseReviewRequestAction, BaseReviewRequestMenuAction]],
    parent_id: Optional[str] = None,
) -> None:
    """Register the given actions as children of the corresponding parent.

    If no parent_id is given, then the actions are assumed to be top-level.

    Deprecated:
        6.0:
        Users should switch to
        :py:const:`reviewboard.actions.actions_registry`. This method will be
        removed in Review Board 8.

    Args:
        actions (iterable of BaseReviewRequestAction):
            The collection of action instances to be registered.

        parent_id (unicode, optional):
            The action ID of the parent of each action instance to be
            registered.

    Raises:
        KeyError:
            The parent action cannot be found or a second registration is
            attempted (action IDs must be unique across all types of actions
            and menu actions, at any depth).

        DepthLimitExceededError:
            The maximum depth limit is exceeded.
    """
    RemovedInReviewBoard90Warning.warn(
        'register_actions has been deprecated and will be removed in '
        'Review Board 9. Please update your code to use '
        'reviewboard.actions.actions_registry.')

    if parent_id:
        parent = actions_registry.get_action(parent_id)
    else:
        parent = None

    for action in actions:
        action.register(parent)


def unregister_actions(
    action_ids: Iterable[str],
) -> None:
    """Unregister each of the actions corresponding to the given IDs.

    Deprecated:
        6.0:
        Users should switch to
        :py:const:`reviewboard.actions.actions_registry`. This method will be
        removed in Review Board 8.

    Args:
        action_ids (iterable of unicode):
            The collection of action IDs corresponding to the actions to be
            removed.

    Raises:
        KeyError:
            An unregistration is attempted before it's registered.
    """
    RemovedInReviewBoard90Warning.warn(
        'unregister_actions has been deprecated and will be removed in '
        'Review Board 9. Please update your code to use '
        'reviewboard.actions.actions_registry.')

    for action_id in action_ids:
        action = actions_registry.get_action(action_id)
        action.unregister()


def clear_all_actions() -> None:
    """Clear all registered actions.

    This method is really only intended to be used by unit tests. We might be
    able to remove this hack once we convert to djblets.registries.

    Deprecated:
        6.0:
        Users should switch to
        :py:const:`reviewboard.actions.actions_registry`. This method will be
        removed in Review Board 8.

    Warning:
        This will clear **all** actions, even if they were registered in
        separate extensions.
    """
    RemovedInReviewBoard90Warning.warn(
        'clear_all_actions has been deprecated and will be removed in '
        'Review Board 9. Please update your code to use '
        'reviewboard.actions.actions_registry.')
    actions_registry.reset()



"""Hosting service support for Forgejo.

Version Added:
    8.0
"""

from __future__ import annotations

from django import forms
from django.utils.translation import gettext_lazy as _

from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm


class ForgejoForm(BaseHostingServiceRepositoryForm):
    """Hosting service form for Forgejo.

    Version Added:
        8.0
    """

    repository_owner = forms.CharField(
        label=_('Repository owner'),
        max_length=128,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The username of the owner of the repository.'),
    )

    repository_name = forms.CharField(
        label=_('Repository name'),
        max_length=128,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository.'),
    )



"""Unit tests for reviewboard.notifications.models"""

from django.urls import reverse

from reviewboard.notifications.models import WebHookTarget
from reviewboard.testing.testcase import TestCase


class WebhookTargetAdminTests(TestCase):
    """Tests for reviewboard.notifications.admin."""

    fixtures = ['test_users', 'test_scmtools']

    def test_webhooktarget_form_redirect(self):
        """Testing that a WebHookTarget form can render on page, and saves
        data correctly
        """
        self.assertTrue(self.client.login(username='admin', password='admin'))
        test_repo = self.create_repository()

        response = self.client.get(
            reverse('admin:notifications_webhooktarget_add'))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse('admin:notifications_webhooktarget_add'),
            {
                'apply_to': WebHookTarget.APPLY_TO_SELECTED_REPOS,
                'enabled': True,
                'encoding': WebHookTarget.ENCODING_JSON,
                'repositories': test_repo.pk,
                'url': 'http://www.google.ca',
            })
        self.assertRedirects(
            response,
            reverse('admin:notifications_webhooktarget_changelist'))
        webhooktarget = WebHookTarget.objects.latest('pk')

        response = self.client.get(
            reverse('admin:notifications_webhooktarget_change',
                    args=(webhooktarget.pk,)))



"""WebHook dispatching logic."""

from __future__ import annotations

import hashlib
import hmac
import logging
from base64 import b64encode
from collections import OrderedDict
from datetime import datetime
from typing import TYPE_CHECKING
from urllib.error import HTTPError
from urllib.parse import urlencode, urlsplit, urlunsplit
from urllib.request import Request

from django.contrib.sites.models import Site
from django.db.models import Model
from django.db.models.query import QuerySet
from django.http.request import HttpRequest
from django.utils.encoding import force_bytes, force_str
from django.utils.functional import cached_property
from django.utils.safestring import SafeText
from django.utils.text import get_text_list
from django.utils.translation import gettext as _
from django.template import Context, Template
from django.template.base import Lexer, Parser
from djblets.log import log_timed
from djblets.siteconfig.models import SiteConfiguration
from djblets.webapi.encoders import (BasicAPIEncoder, JSONEncoderAdapter,
                                     ResourceAPIEncoder, XMLEncoderAdapter)

from reviewboard import get_package_version
from reviewboard.certs.http import urlopen
from reviewboard.notifications.models import WebHookTarget
from reviewboard.reviews.models import Review, ReviewRequest
from reviewboard.reviews.signals import (review_request_closed,
                                         review_request_published,
                                         review_request_reopened,
                                         review_published,
                                         reply_published)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from typelets.json import JSONDict


logger = logging.getLogger(__name__)


class FakeHTTPRequest(HttpRequest):
    """A fake HttpRequest implementation.

    The WebAPI serialization methods use HttpRequest.build_absolute_uri to
    generate all the links, but none of the various signals that generate
    webhook events have the request plumbed through. Since we don't actually
    need a valid request, this impersonates it enough to get valid results from
    build_absolute_uri.
    """

    def __init__(self, user, local_site_name=None):
        """Initialize a FakeHTTPRequest.

        Args:
            user (django.contrib.auth.models.User):
                The user who initiated the request.

            local_site_name (unicode, optional):
                The local site name (if the request was carried out against a
                local site).
        """
        super(FakeHTTPRequest, self).__init__()

        self.user = user
        self._local_site_name = local_site_name
        self._host = Site.objects.get_current().domain

    @cached_property
    def scheme(self):
        """The protocol scheme used to access the web server.

        This is needed by the underlying class to build absolute URLs. It
        returns the scheme from the current site configuration.

        Version Added:
            4.0.3

        Type:
            unicode
        """
        siteconfig = SiteConfiguration.objects.get_current()

        return siteconfig.get('site_domain_method')

    def get_host(self):
        """Return the hostname for the server.

        This is needed by the underlying class to build absolute URLs. It
        returns the hostname from the configured site information.

        Returns:
            unicode:
            The hostname.
        """
        return self._host


class CustomPayloadParser(Parser):
    """A custom template parser that blocks certain tags.

    This extends Django's Parser class for template parsing, and removes
    some built-in tags, in order to prevent mailicious use.
    """
    BLACKLISTED_TAGS = ('block', 'debug', 'extends', 'include', 'load', 'ssi')

    def __init__(self, *args, **kwargs):
        super(CustomPayloadParser, self).__init__(*args, **kwargs)

        # Remove some built-in tags that we don't want to expose.
        # There are no built-in filters we have to worry about.
        for tag_name in self.BLACKLISTED_TAGS:
            try:
                del self.tags[tag_name]
            except KeyError:
                pass

    def invalid_block_tag(self, token, command, parse_until=None):
        """Raise an error when an invalid block tag is found.

        Normally Django produces a suitable error, but in modern versions of
        Django, the error is _too_ helpful, reminding the user to register or
        load the tag. This isn't useful for WebHooks, so we override to use
        the older, simpler message used in Django 1.6.

        Args:
            token (django.template.base.Token):
                The token representing the block tag.

            command (unicode):
                The name of the block tag that was found.

            parse_until (list of django.template.base.Token, optional):
                The list of tokens that were expected to be parsed instead
                of this token.
        """
        if parse_until:
            raise self.error(
                token,
                _("Invalid block tag: '%(found_tag)s', expected "
                  "%(expected_tag)s")
                % {
                    'found_tag': command,
                    'expected_tag': get_text_list([
                        "'%s'" % tag
                        for tag in parse_until
                    ]),
                })
        else:
            raise self.error(token, "Invalid block tag: '%s'" % command)


def render_custom_content(body, context_data={}):
    """Render custom content for the payload using Django templating.

    This will take the custom payload content template provided by
    the user and render it using a stripped down version of Django's
    templating system.

    In order to keep the payload safe, we use a limited Context along with a
    custom Parser that blocks certain template tags. This gives us
    tags like ``{% for %}`` and ``{% if %}``, but blacklists tags like
    ``{% load %}`` and ``{% include %}``.

    Args:
        body (unicode):
            The template content to render.

        context_data (dict, optional):
            Context data for the template.

    Returns:
        unicode:
        The rendered template.

    Raises:
        django.template.TemplateSyntaxError:
            There was a syntax error in the template.
    """
    template = Template('')
    lexer = Lexer(body)
    parser_args = (template.engine.template_libraries,
                   template.engine.template_builtins,
                   template.origin)

    parser = CustomPayloadParser(lexer.tokenize(), *parser_args)
    template.nodelist = parser.parse()

    return template.render(Context(context_data))


def normalize_webhook_payload(payload, request, use_string_keys=False):
    """Normalize a payload for a WebHook, returning a safe, primitive version.

    This will take a payload containing various data types and model references
    and turn it into a payload built out of specific, whitelisted types
    (strings, bools, ints, dicts, lists, and datetimes). This payload is
    safe to include in custom templates without worrying about access to
    dangerous functions, and is easy to serialize.

    Args:
        payload (dict):
            The payload to normalize.

        request (django.http.HttpRequest):
            The HTTP request from the client.

        use_string_keys (bool, optional):
            Whether to normalize all keys to strings.

    Returns:
        dict:
        The normalized payload.

    Raises:
        TypeError:
            An unsupported data type was found in the payload. This is an
            issue with the caller.
    """
    def _normalize_key(key):
        if key is None:
            if use_string_keys:
                return 'null'

            return None
        elif isinstance(key, str):
            return key
        elif isinstance(key, (SafeText, bool, float)):
            return str(key)
        elif isinstance(key, bytes):
            return force_str(key)
        elif isinstance(key, int):
            if use_string_keys:
                return force_str(key)

            return key
        else:
            raise TypeError(
                _('%s is not a valid data type for dictionary keys in '
                  'WebHook payloads.')
                % type(key))

    def _normalize_value(value):
        if value is None:
            return None

        if isinstance(value, SafeText):
            return str(value)
        elif isinstance(value, bytes):
            return force_str(value)
        elif isinstance(value, (bool, datetime, float, str, int)):
            return value
        elif isinstance(value, dict):
            return OrderedDict(
                (_normalize_key(dict_key), _normalize_value(dict_value))
                for dict_key, dict_value in value.items()
            )
        elif isinstance(value, (list, tuple)):
            return [
                _normalize_value(item)
                for item in value
            ]
        elif isinstance(value, (Model, QuerySet)):
            result = resource_encoder.encode(value, request=request)

            if result is not None:
                return _normalize_value(result)

        raise TypeError(
            _('%s is not a valid data type for values in WebHook payloads.')
            % type(value))

    resource_encoder = ResourceAPIEncoder()

    return _normalize_value(payload)


def dispatch_webhook_event(
    request: HttpRequest,
    webhook_targets: Sequence[WebHookTarget],
    event: str,
    payload: JSONDict,
) -> None:
    """Dispatch the given event and payload to the given WebHook targets.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        webhook_targets (list of
                         reviewboard.notifications.models.WebHookTarget):
            The list of WebHook targets containing endpoint URLs to dispatch
            to.

        event (str):
            The name of the event being dispatched.

        payload (dict):
            The payload data to encode for the WebHook payload.

    Raises:
        ValueError:
            There was an error with the payload format. Details are in the
            log and the exception message.
    """
    encoder = BasicAPIEncoder()
    bodies = {}

    raw_norm_payload = None
    json_norm_payload = None

    for webhook_target in webhook_targets:
        use_custom_content = webhook_target.use_custom_content
        encoding = webhook_target.encoding

        # See how we need to handle normalizing this payload. If we need
        # something JSON-safe, then we need to go the more aggressive route
        # and normalize keys to strings.
        if raw_norm_payload is None or json_norm_payload is None:
            try:
                if (raw_norm_payload is None and
                    (use_custom_content or
                     encoding == webhook_target.ENCODING_XML)):
                    # This payload's going to be provided for XML and custom
                    # templates. We don't want to alter the keys at all.
                    raw_norm_payload = normalize_webhook_payload(
                        payload=payload,
                        request=request)
                elif (json_norm_payload is None and
                      not use_custom_content and
                      encoding in (webhook_target.ENCODING_JSON,
                                   webhook_target.ENCODING_FORM_DATA)):
                    # This payload's going to be provided for JSON or
                    # form-data. We want to normalize all keys to strings.
                    json_norm_payload = normalize_webhook_payload(
                        payload=payload,
                        request=request,
                        use_string_keys=True)
            except TypeError as e:
                logger.exception('WebHook payload passed to '
                                 'dispatch_webhook_event containing invalid '
                                 'data types: %s',
                                 e,
                                 extra={'request': request})

                raise ValueError(str(e))

        if use_custom_content:
            try:
                assert raw_norm_payload is not None
                body = render_custom_content(webhook_target.custom_content,
                                             raw_norm_payload)
                body = force_bytes(body)
            except Exception as e:
                logger.exception('Could not render WebHook payload: %s', e,
                                 extra={'request': request})
                continue
        else:
            if encoding not in bodies:
                try:
                    if encoding == webhook_target.ENCODING_JSON:
                        assert json_norm_payload is not None
                        adapter = JSONEncoderAdapter(encoder)
                        body = adapter.encode(json_norm_payload,
                                              request=request)
                    elif encoding == webhook_target.ENCODING_XML:
                        assert raw_norm_payload is not None
                        adapter = XMLEncoderAdapter(encoder)
                        body = adapter.encode(raw_norm_payload,
                                              request=request)
                    elif encoding == webhook_target.ENCODING_FORM_DATA:
                        assert json_norm_payload is not None
                        adapter = JSONEncoderAdapter(encoder)
                        body = urlencode({
                            'payload': adapter.encode(json_norm_payload,
                                                      request=request),
                        })
                    else:
                        logger.error('Unexpected WebHookTarget encoding "%s" '
                                     'for ID %s',
                                     encoding, webhook_target.pk,
                                     extra={'request': request})
                        continue
                except Exception as e:
                    logger.exception('Could not encode WebHook payload: %s',
                                     e,
                                     extra={'request': request})
                    continue

                body = force_bytes(body)
                bodies[encoding] = body
            else:
                body = bodies[encoding]

        headers = {
            'X-ReviewBoard-Event': event,
            'Content-Type': webhook_target.encoding,
            'Content-Length': '%s' % len(body),
            'User-Agent': 'ReviewBoard-WebHook/%s' % get_package_version(),
        }

        if webhook_target.secret:
            signer = hmac.new(webhook_target.secret.encode('utf-8'), body,
                              hashlib.sha1)
            headers['X-Hub-Signature'] = 'sha1=%s' % signer.hexdigest()

        with log_timed(f'Dispatching webhook for event {event} to '
                       f'{webhook_target.url}',
                       logger=logger,
                       request=request) as log_timer:
            try:
                url = webhook_target.url
                url_parts = urlsplit(url)

                if url_parts.username or url_parts.password:
                    credentials, netloc = url_parts.netloc.split('@', 1)
                    url = urlunsplit(
                        (url_parts.scheme, netloc, url_parts.path,
                         url_parts.query, url_parts.fragment))
                    headers['Authorization'] = \
                        'Basic %s' % b64encode(credentials.encode('utf-8'))

                urlopen(
                    Request(url, body, headers),
                    local_site=webhook_target.local_site,
                )
            except Exception as e:
                logger.exception('[%s] Could not dispatch WebHook to %s: %s',
                                 log_timer.trace_id, webhook_target.url, e,
                                 extra={'request': request})

                if isinstance(e, HTTPError):
                    logger.info('[%s] Error response from %s: %s %s\n%s',
                                log_timer.trace_id, webhook_target.url,
                                e.code, e.reason, e.read(),
                                extra={'request': request})


def _serialize_review(review, request):
    return {
        'review_request': review.review_request,
        'review': review,
        'diff_comments': review.comments.all(),
        'file_attachment_comments': review.file_attachment_comments.all(),
        'screenshot_comments': review.screenshot_comments.all(),
        'general_comments': review.general_comments.all(),
    }


def _serialize_reply(reply, request):
    return {
        'review_request': reply.review_request,
        'reply': reply,
        'diff_comments': reply.comments.all(),
        'file_attachment_comments': reply.file_attachment_comments.all(),
        'screenshot_comments': reply.screenshot_comments.all(),
        'general_comments': reply.general_comments.all(),
    }


def review_request_closed_cb(user, review_request, close_type, **kwargs):
    event = 'review_request_closed'
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        if close_type == review_request.SUBMITTED:
            close_type = 'submitted'
        elif close_type == review_request.DISCARDED:
            close_type = 'discarded'
        else:
            logger.error('Unexpected close type %s for review request %s '
                         'when dispatching webhook.',
                         type, review_request.pk)
            return

        if not user:
            user = review_request.submitter

        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = {
            'event': event,
            'closed_by': user,
            'close_type': close_type,
            'review_request': review_request,
        }

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def review_request_published_cb(user, review_request, changedesc,
                                **kwargs):
    event = 'review_request_published'
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = {
            'event': event,
            'is_new': changedesc is None,
            'review_request': review_request,
        }

        if changedesc:
            payload['change'] = changedesc

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def review_request_reopened_cb(user, review_request, **kwargs):
    event = 'review_request_reopened'
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        if not user:
            user = review_request.submitter

        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = {
            'event': event,
            'reopened_by': user,
            'review_request': review_request,
        }

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def review_published_cb(user, review, **kwargs):
    event = 'review_published'
    review_request = review.review_request
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = _serialize_review(review, request)
        payload['event'] = event

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def reply_published_cb(user, reply, **kwargs):
    event = 'reply_published'
    review_request = reply.review_request
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = _serialize_reply(reply, request)
        payload['event'] = event

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def connect_signals():
    review_request_closed.connect(review_request_closed_cb,
                                  sender=ReviewRequest)
    review_request_published.connect(review_request_published_cb,
                                     sender=ReviewRequest)
    review_request_reopened.connect(review_request_reopened_cb,
                                    sender=ReviewRequest)

    review_published.connect(review_published_cb, sender=Review)
    reply_published.connect(reply_published_cb, sender=Review)



"""Unit tests for SAML settings.

Version Added:
    7.0.4
"""

from __future__ import annotations

from reviewboard.testing import TestCase

from reviewboard.accounts.sso.backends.saml.settings import get_saml2_settings


class SAMLSettingsTests(TestCase):
    """Unit tests for SAML settings.

    Version Added:
        7.0.4
    """

    def test_advanced_config(self) -> None:
        """Testing get_saml2_settings with advanced config override"""
        django_settings = {
            'SAML_CONFIG_ADVANCED': {
                'security': {
                    'requestedAuthnContext': False,
                },
            },
        }

        with self.settings(**django_settings):
            saml_settings = get_saml2_settings()

            self.assertEqual(saml_settings['security'], {
                'requestedAuthnContext': False,
            })



"""Base classes for SSO views.

Version Added:
    5.0
"""

import logging

from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from django.views.generic.base import View


logger = logging.getLogger(__name__)


@method_decorator([sensitive_post_parameters(), never_cache],
                  name='dispatch')
class BaseSSOView(View):
    """Base class for SSO views.

    Version Added:
        5.0
    """

    #: The SSO backend.
    #:
    #: Type:
    #:     reviewboard.accounts.sso.backends.SSOBackend
    sso_backend = None



"""Unit tests for the All Review Requests page.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, TYPE_CHECKING

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.db.models import Count, Q
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import Profile
from reviewboard.datagrids.tests.base import BaseViewTestCase
from reviewboard.reviews.models import ReviewRequest
from reviewboard.reviews.testing.queries.review_requests import \
    get_review_requests_accessible_q
from reviewboard.site.models import LocalSite
from reviewboard.testing.queries.http import get_http_request_start_equeries

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries


class AllReviewRequestViewTests(BaseViewTestCase):
    """Unit tests for the all_review_requests view."""

    datagrid_url = '/r/'

    @add_fixtures(['test_users'])
    def test_with_access(self):
        """Testing all_review_requests view"""
        self._test_with_access()

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_local_site(self):
        """Testing all_review_requests view on a Local Site"""
        self._test_with_access(with_local_site=True,
                               local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_local_sites_in_db(self):
        """Testing all_review_requests view with Local Sites in the database
        """
        self._test_with_access(local_sites_in_db=True)

    def test_as_anonymous_and_redirect(self):
        """Testing all_review_requests view as anonymous user
        with anonymous access disabled
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': True},
                                      reload_settings=False):
            response = self.client.get('/r/')

        self.assertEqual(response.status_code, 302)

    @add_fixtures(['test_scmtools', 'test_users'])
    def test_with_private_review_requests(self):
        """Testing all_review_requests view with private review requests"""
        self._test_with_private_review_requests()

    @add_fixtures(['test_site', 'test_scmtools', 'test_users'])
    def test_with_private_review_requests_with_local_site(self):
        """Testing all_review_requests view with private review requests on
        a Local Site
        """
        self._test_with_private_review_requests(with_local_site=True,
                                                local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_scmtools', 'test_users'])
    def test_with_private_review_requests_with_local_sites_in_db(self):
        """Testing all_review_requests view with private review requests with
        Local Sites in the database
        """
        self._test_with_private_review_requests(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_with_inactive_users(self):
        """Testing all_review_requests view with review requests from inactive
        users
        """
        self._test_with_inactive_users()

    @add_fixtures(['test_site', 'test_users'])
    def test_with_inactive_users_with_local_site(self):
        """Testing all_review_requests view with review requests from inactive
        users on a Local Site
        """
        self._test_with_inactive_users(with_local_site=True,
                                       local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_with_inactive_users_with_local_sites_in_db(self):
        """Testing all_review_requests view with review requests from inactive
        users with Local Sites in the database
        """
        self._test_with_inactive_users(local_sites_in_db=True)

    def _test_with_access(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for a user with access to the datagrid.

        This accesses the All Review Requests datagrid as a logged-in user with
        access to the global site or Local Site (depending on the test),
        checking queries and datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        review_request_1 = self.create_review_request(
            local_site=local_site,
            local_id=1,
            summary='Test 1',
            publish=True)
        review_request_2 = self.create_review_request(
            local_site=local_site,
            local_id=2,
            summary='Test 2',
            publish=True)
        review_request_3 = self.create_review_request(
            local_site=local_site,
            local_id=3,
            summary='Test 3',
            publish=True)

        self.client.login(username='grumpy', password='grumpy')
        user = User.objects.get(username='grumpy')
        profile = user.get_profile()

        if local_site:
            local_site.users.add(user)

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=[
                review_request_3.pk,
                review_request_2.pk,
                review_request_1.pk,
            ])

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site))

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 3)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 1')

    def _test_with_private_review_requests(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for checking access to private review requests.

        This accesses the All Review Requests datagrid as a logged-in user with
        access to the global site or Local Site (depending on the test),
        checking queries and datagrid row results to make sure only accessible
        private review requests are included.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        user = User.objects.get(username='grumpy')
        profile = user.get_profile()

        if local_site:
            local_site.users.add(user)

        # These are public
        review_request_1 = self.create_review_request(
            local_site=local_site,
            local_id=1,
            summary='Test 1',
            publish=True)
        review_request_2 = self.create_review_request(
            local_site=local_site,
            local_id=2,
            summary='Test 2',
            publish=True)

        repository1 = self.create_repository(
            local_site=local_site,
            name='repo1',
            public=False,
            users=[user])
        review_request_3 = self.create_review_request(
            local_site=local_site,
            local_id=3,
            summary='Test 3',
            repository=repository1,
            publish=True)

        group1 = self.create_review_group(
            local_site=local_site,
            name='group1',
            invite_only=True,
            users=[user])
        review_request_4 = self.create_review_request(
            local_site=local_site,
            local_id=4,
            summary='Test 4',
            publish=True,
            target_groups=[group1])

        # These are private
        repository2 = self.create_repository(
            local_site=local_site,
            name='group2',
            public=False)
        self.create_review_request(
            local_site=local_site,
            local_id=5,
            summary='Test 5',
            repository=repository2,
            publish=True)

        group2 = self.create_review_group(
            local_site=local_site,
            name='group2',
            invite_only=True)
        self.create_review_request(
            local_site=local_site,
            local_id=6,
            summary='Test 6',
            publish=True,
            target_groups=[group2])

        # Log in and check what we get.
        self.client.login(username='grumpy', password='grumpy')

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=[
                review_request_4.pk,
                review_request_3.pk,
                review_request_2.pk,
                review_request_1.pk,
            ],
            repositories_pks=[repository1.pk],
            target_groups_pks=[group1.pk])

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site))

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 4)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 4')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 3')
        self.assertEqual(datagrid.rows[2]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[3]['object'].summary, 'Test 1')

    def _test_with_inactive_users(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for checking for review requests from inactive users.

        This accesses the All Review Requests datagrid as a logged-in user with
        access to the global site or Local Site (depending on the test),
        checking queries and datagrid row results to make sure review requests
        from inactive users are included.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        User.objects.filter(username='doc').update(is_active=False)

        review_request_1 = self.create_review_request(
            local_site=local_site,
            local_id=1,
            summary='Test 1',
            submitter='doc',
            publish=True)
        review_request_1.close(ReviewRequest.SUBMITTED)

        review_request_2 = self.create_review_request(
            local_site=local_site,
            local_id=2,
            summary='Test 2',
            submitter='grumpy',
            publish=True)

        self.client.login(username='grumpy', password='grumpy')
        user = User.objects.get(username='grumpy')
        profile = user.get_profile()

        if local_site:
            local_site.users.add(user)

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_request_pks=[
                review_request_2.pk,
                review_request_1.pk,
            ])

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site))

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 2)
        self.assertEqual(datagrid.rows[0]['object'].summary, 'Test 2')
        self.assertEqual(datagrid.rows[1]['object'].summary, 'Test 1')

    def _build_datagrid_equeries(
        self,
        *,
        user: User,
        profile: Profile,
        review_request_pks: List[int],
        local_site: Optional[LocalSite] = None,
        local_sites_in_db: bool = False,
        repositories_pks: List[int] = [],
        target_groups_pks: List[int] = [],
    ) -> ExpectedQueries:
        """Return expected queries for viewing the datagrid.

        This assumes that the user has access to the datagrid, and that any
        cacheable state is already cached.

        Version Added:
            5.0.7

        Args:
            user (django.contrib.auth.models.User):
                The user accessing the datagrid.

            profile (reviewboard.accounts.models.Profile):
                The user's profile.

            review_request_pks (list of int):
                The list of review request IDs that are expected to be
                listed, in result order.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site that's being accessed, if any.

            local_sites_in_db (bool, optional):
                Whether the database contains any Local Sites.

            repositories_pks (list of int, optional):
                The list of any repository IDs that should be part of the
                query.

            target_groups_pks (list of int, optional):
                The list of any target group IDs that should be part of the
                query.

        Returns:
            list of dict:
            The list of expected queries.
        """
        user_pk = user.pk
        extra: Dict[str, Any] = {
            'new_review_count': (f"""
                SELECT COUNT(*)
                  FROM reviews_review, accounts_reviewrequestvisit
                  WHERE reviews_review.public
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
                    AND accounts_reviewrequestvisit.review_request_id =
                        reviews_reviewrequest.id
                    AND accounts_reviewrequestvisit.user_id = {user_pk}
                    AND reviews_review.timestamp >
                        accounts_reviewrequestvisit.timestamp
                    AND reviews_review.user_id != {user_pk}
            """, []),
            'draft_summary': ("""
                SELECT reviews_reviewrequestdraft.summary
                  FROM reviews_reviewrequestdraft
                  WHERE reviews_reviewrequestdraft.review_request_id =
                        reviews_reviewrequest.id
            """, []),
            'visibility': (f"""
                SELECT accounts_reviewrequestvisit.visibility
                  FROM accounts_reviewrequestvisit
                 WHERE accounts_reviewrequestvisit.review_request_id =
                       reviews_reviewrequest.id
                   AND accounts_reviewrequestvisit.user_id = {user_pk}
            """, []),
        }

        rows_q_result = get_review_requests_accessible_q(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=local_sites_in_db,
            filter_private=True,
            show_inactive=True,
            status=None,
            accessible_repository_ids=repositories_pks,
            accessible_review_group_ids=target_groups_pks,
            needs_local_site_profile_query=True)
        rows_q = rows_q_result['q']
        rows_q_tables = rows_q_result['tables']
        rows_q_num_joins = len(rows_q_tables) - 1
        rows_q_join_types = rows_q_result.get('join_types', {})
        rows_q_subqueries = rows_q_result.get('subqueries', [])

        equeries = get_http_request_start_equeries(
            user=user,
            local_site=local_site)
        equeries += rows_q_result.get('prep_equeries', [])
        equeries += [
            {
                '__note__': 'Update datagrid state on the user profile',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=profile.pk),
            },
            {
                '__note__': (
                    'Fetch the number of items across all datagrid pages'
                ),
                'annotations': {'__count': Count('*')},
                'join_types': rows_q_join_types,
                'model': ReviewRequest,
                'num_joins': rows_q_num_joins,
                'subqueries': rows_q_subqueries,
                'tables': rows_q_tables,
                'where': (Q(rows_q) &
                          Q(local_site=local_site)),
            },
            {
                '__note__': 'Fetch the IDs of the items for one page',
                'distinct': True,
                'extra': extra,
                'join_types': rows_q_join_types,
                'limit': len(review_request_pks),
                'model': ReviewRequest,
                'num_joins': rows_q_num_joins,
                'order_by': ('-last_updated',),
                'subqueries': rows_q_subqueries,
                'tables': rows_q_tables,
                'values_select': ('pk',),
                'where': (Q(rows_q) &
                          Q(local_site=local_site)),
            },
            {
                '__note__': (
                    "Fetch the IDs of the page's review requests that are "
                    "starred."
                ),
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'values_select': ('pk',),
                'where': (Q(starred_by__id=profile.pk) &
                          Q(pk__in=review_request_pks)),
            },
        ]

        if local_site is not None:
            equeries += [
                {
                    '__note__': (
                        'Fetch the data for one page based on the IDs on the '
                        'Local Site'
                    ),
                    'extra': extra,
                    'model': ReviewRequest,
                    'select_related': {
                        'local_site',
                        'submitter',
                    },
                    'where': Q(pk__in=review_request_pks),
                },
            ]
        else:
            equeries += [
                {
                    '__note__': 'Fetch the data for one page based on the IDs',
                    'extra': extra,
                    'model': ReviewRequest,
                    'select_related': {'submitter'},
                    'where': Q(pk__in=review_request_pks),
                },
            ]

        return equeries



"""Utilities for dealing with DiffCommits."""

from __future__ import annotations

import base64
import json
from itertools import chain
from typing import Any, Optional, TYPE_CHECKING

from django.utils.encoding import force_bytes
from typing_extensions import NotRequired, TypedDict

from reviewboard.scmtools.core import PRE_CREATION, UNKNOWN

if TYPE_CHECKING:
    from reviewboard.scmtools.core import FileLookupContext
    from reviewboard.scmtools.models import Repository


def get_file_exists_in_history(
    validation_info: dict[str, Any],
    repository: Repository,
    parent_id: str,
    path: str,
    revision: str,
    base_commit_id: Optional[str] = None,
    *,
    context: Optional[FileLookupContext] = None,
    **kwargs,
) -> bool:
    """Return whether or not the file exists, given the validation information.

    This will walk through the commit chain in ``validation_info``, starting
    with ``parent_id``, looking for a file entry that matches the given
    ``path`` and (by default) ``revision``. If found, the file is considered
    to exist. If not found, it will fall back to checking the repository.

    This can also operate in a loose validation mode. In this mode, only
    file paths are compared, not revisions. This is required for diffs that
    don't provide per-file revision information (such as Mercurial's plain
    and Git-style diffs). If a match is found, the associated commit ID is
    stored as ``context.file_extra_data['__validated_parent_id']``, allowing
    for post-processing of the source revision to occur.

    Strict validation mode is the default. Loose validation must be opted into
    by setting ``context.diff_extra_data['has_per_file_revisions'] = False``.
    This is done automatically when a
    :py:class:`~reviewboard.diffviewer.parser.BaseDiffParser` subclass sets
    :py:attr:`has_per_file_revisions
    <reviewboard.diffviewer.parser.BaseDiffParser.has_per_file_revisions>`
    to ``False``.

    Version Changed:
        7.0.2:
        * Added the ``context`` argument.
        * Added the loose validation mode.

    Version Changed:
        4.0.5:
        Removed explicit arguments for ``base_commit_id`` and ``request``, and
        added ``**kwargs``.

    Args:
        validation_info (dict):
            Validation metadata generated by the
            :py:class:`~reviewboard.webapi.resources.validate_diffcommit.
            ValidateDiffCommitResource`.

        repository (reviewboard.scmtools.models.Repository):
            The repository.

        parent_id (str):
            The parent commit ID of the commit currently being processed.

        path (str):
            The file path.

        revision (str):
            The revision of the file to retrieve.

        context (reviewboard.scmtools.core.FileLookupContext, optional):
            The file lookup context used to validate this repository.

            Version Added:
                7.0.2

        **kwargs (dict):
            Additional keyword arguments normally expected by
            :py:meth:`Repository.get_file_exists
            <reviewboard.scmtools.models.Repository.get_file_exists>`.

            Version Added:
                4.0.5

    Returns:
        bool:
        Whether or not the file exists.
    """
    match_parent_revisions = (
        context is None or
        context.diff_extra_data.get('has_per_file_revisions', True))

    while parent_id in validation_info:
        entry = validation_info[parent_id]
        tree = entry['tree']

        if revision == UNKNOWN:
            for removed_info in tree['removed']:
                if removed_info['filename'] == path:
                    return False

        for change_info in chain(tree['added'], tree['modified']):
            if change_info['filename'] == path:
                if revision == UNKNOWN:
                    return True

                if match_parent_revisions:
                    # In the standard case, we have per-file revisions, and
                    # can use that to get a specific match within the
                    # validation history.
                    if change_info['revision'] == revision:
                        return True
                else:
                    # In a more limited case, we may not know the per-file
                    # revision, and instead have to limit our scan to
                    # the nearest filename. This is the case with Mercurial
                    # diffs.
                    #
                    # We'll be recording what we found, temporarily. This
                    # will be used to update the source filename of a
                    # generated FileDiff.
                    assert context is not None

                    context.file_extra_data['__validated_parent_id'] = \
                        parent_id

                    return True

        parent_id = entry['parent_id']

    # We did not find an entry in our validation info, so we need to fall back
    # to checking the repository.
    return repository.get_file_exists(path=path,
                                      revision=revision,
                                      base_commit_id=base_commit_id,
                                      context=context,
                                      **kwargs)


def exclude_ancestor_filediffs(to_filter, all_filediffs=None):
    """Exclude all ancestor FileDiffs from the given list and return the rest.

    A :pyclass:`~reviewboard.diffviewer.models.filediff.FileDiff` is considered
    an ancestor of another if it occurs in a previous commit and modifies the
    same file.

    As a result, only the most recent (commit-wise) FileDiffs that modify a
    given file will be included in the result.

    Args:
        to_filter (list of reviewboard.diffviewer.models.filediff.FileDiff):
            The FileDiffs to filter.

        all_filediffs (list of reviewboard.diffviewer.models.filediff.FileDiff,
                       optional):
            The list of all FileDiffs in the :py:class:`~reviewboard.
            diffviewer.models.diffset.DiffSet>`.

            If not provided, it is assumed that ``to_filter`` is a list of all
            FileDiffs in the :py:class:`~reviewboard.diffviewer.models.
            diffset.DiffSet>`.

    Returns:
        list of reviewboard.diffviewer.models.filediff.FileDiff:
        The FileDiffs that are not ancestors of other FileDiffs.
    """
    if all_filediffs is None:
        all_filediffs = to_filter

    ancestor_pks = {
        ancestor.pk
        for filediff in to_filter
        for ancestor in filediff.get_ancestors(minimal=False,
                                               filediffs=all_filediffs)
    }

    return [
        filediff
        for filediff in to_filter
        if filediff.pk not in ancestor_pks
    ]


def deserialize_validation_info(raw):
    """Deserialize the raw validation info.

    Args:
        raw (unicode or bytes):
            The raw validation info from the client.

    Returns:
        dict:
        The deserialized validation info.

    Raises:
        ValueError:
            Either the data could not be base64-decoded or the resulting JSON
            was of an invalid format (i.e., it was not a dictionary).

        TypeError:
            The base64-decoded data could not be interpreted as JSON.
    """
    value = json.loads(base64.b64decode(force_bytes(raw)).decode('utf-8'))

    if not isinstance(value, dict):
        raise ValueError('Invalid format.')

    return value


def serialize_validation_info(info):
    """Serialize the given validation info into a raw format.

    Args:
        info (dict):
            The dictionary of validation info.

    Returns:
        unicode:
        The base64-encoded JSON of the validation info.
    """
    data = json.dumps(info).encode('utf-8')

    return base64.b64encode(data).decode('utf-8')


def update_validation_info(validation_info, commit_id, parent_id, filediffs):
    """Update the validation info with a new commit.

    Args:
        validation_info (dict):
            The dictionary of validation info. This will be modified in-place.

            This is a mapping of commit IDs to their metadata. Each metadata
            dictionary contains the following keys:

            ``parent_id``:
                The commit ID of the parent commit.

            ``tree``:
                A dictionary of the added, removed, and modified files in this
                commit.

        commit_id (unicode):
            The commit ID of the commit whose metadata is being added to the
            dictionary.

            This must not already be present in ``validation_info``.

        parent_id (unicode):
            The commit ID of the parent commit.

            This must be present in ``validation_info`` *unless* this is the
            first commit being added (i.e., ``validation_info`` is empty).

        filediffs (list of reviewboard.diffviewer.models.filediff.FileDiff):
            The parsed FileDiffs from :py:func:`~reviewboard.diffviewer.
            filediff_creator.create_filediffs`.

    Returns:
        dict:
        The dictionary of validation info.
    """
    from reviewboard.diffviewer.models import FileDiff

    assert validation_info == {} or parent_id in validation_info
    assert commit_id not in validation_info

    added = []
    removed = []
    modified = []

    for f in filediffs:
        if f.status in (FileDiff.DELETED, FileDiff.MOVED):
            removed.append({
                'filename': f.source_file,
                'revision': f.source_revision,
            })

        if (f.status in (FileDiff.COPIED, FileDiff.MOVED) or
            (f.status == FileDiff.MODIFIED and
             f.source_revision == PRE_CREATION)):
            added.append({
                'filename': f.dest_file,
                'revision': f.dest_detail,
            })
        elif f.status == FileDiff.MODIFIED:
            modified.append({
                'filename': f.dest_file,
                'revision': f.dest_detail,
            })

    validation_info[commit_id] = {
        'parent_id': parent_id,
        'tree': {
            'added': added,
            'modified': modified,
            'removed': removed,
        },
    }

    return validation_info


class SerializedCommitHistoryDiffEntry(TypedDict):
    """Serialized version of a CommitHistoryDiffEntry.

    Version Added:
        7.0
    """

    #: The type of change in the commit history diff.
    #:
    #: This will be one of "added", "removed", "modified", or "unmodified".
    entry_type: str

    #: The new ID of the commit in the diff line.
    new_commit_id: NotRequired[int]

    #: The old ID of the commit in the diff line.
    old_commit_id: NotRequired[int]


class CommitHistoryDiffEntry(object):
    """An entry in a commit history diff."""

    COMMIT_ADDED = 'added'
    COMMIT_REMOVED = 'removed'
    COMMIT_MODIFIED = 'modified'
    COMMIT_UNMODIFIED = 'unmodified'

    entry_types = (
        COMMIT_ADDED,
        COMMIT_REMOVED,
        COMMIT_MODIFIED,
        COMMIT_UNMODIFIED,
    )

    def __init__(self, entry_type, old_commit=None, new_commit=None):
        """Initialize the CommitHistoryDiffEntry object.

        Args:
            entry_type (unicode):
                The commit type. This must be one of the values in
                :py:attr:`entry_types`.

            old_commit (reviewboard.diffviewer.models.diffcommit.DiffCommit,
                        optional):
                The old commit. This is required if the commit type is one of:

                * :py:data:`COMMIT_REMOVED`
                * :py:data:`COMMIT_MODIFIED`
                * :py:data:`COMMIT_UNMODIFIED`

            new_commit (reviewboard.diffviewer.models.diffcommit.DiffCommit,
                        optional):
                The new commit. This is required if the commit type is one of:

                * :py:data:`COMMIT_ADDED`
                * :py:data:`COMMIT_MODIFIED`
                * :py:data:`COMMIT_UNMODIFIED`

        Raises:
            ValueError:
                The value of ``entry_type`` was invalid or the wrong commits
                were specified.
        """
        if entry_type not in self.entry_types:
            raise ValueError('Invalid entry_type: "%s"' % entry_type)

        if not old_commit and entry_type != self.COMMIT_ADDED:
            raise ValueError('old_commit required for given commit type.')

        if not new_commit and entry_type != self.COMMIT_REMOVED:
            raise ValueError('new_commit required for given commit type')

        self.entry_type = entry_type
        self.old_commit = old_commit
        self.new_commit = new_commit

    def serialize(self) -> SerializedCommitHistoryDiffEntry:
        """Serialize the entry to a dictionary.

        Returns:
            dict:
            A dictionary of the serialized information.
        """
        result: SerializedCommitHistoryDiffEntry = {
            'entry_type': self.entry_type,
        }

        if self.new_commit:
            result['new_commit_id'] = self.new_commit.pk

        if self.old_commit:
            result['old_commit_id'] = self.old_commit.pk

        return result

    def __eq__(self, other):
        """Compare two entries for equality.

        Two entries are equal if and only if their attributes match.

        Args:
            other (CommitHistoryDiffEntry):
                The entry to compare against.

        Returns:
            bool:
            Whether or not this entry and the other entry are equal.
        """
        return (self.entry_type == other.entry_type,
                self.old_commit == other.old_commit,
                self.new_commit == other.new_commit)

    def __ne__(self, other):
        """Compare two entries for inequality.

        Two entries are not equal if and only if any of their attributes don't
        match.

        Args:
            other (CommitHistoryDiffEntry):
                The entry to compare against.

        Returns:
            bool:
            Whether or not this entry and the other entry are not equal.
        """
        return not self == other

    def __repr__(self):
        """Return a string representation of the entry.

        Returns:
            unicode:
            A string representation of the entry.
        """
        return (
            '<CommitHistoryDiffEntry(entry_type=%s, '
            'old_commit=%s, new_commit=%s)>'
            % (self.entry_type, self.old_commit, self.new_commit)
        )


def diff_histories(old_history, new_history):
    """Yield the entries in the diff between the old and new histories.

    Args:
        old_history (list of reviewboard.diffviewer.models.diffcommit.
                     DiffCommit):
            The list of commits from a previous
            :py:class:`~reviewboard.diffviewer.models.diffset.DiffSet`.

        new_history (list of reviewboard.diffviewer.models.diffcommit.
                     DiffCommit):
            The list of commits from the new
            :py:class:`~reviewboard.diffviewer.models.diffset.DiffSet`.

    Yields:
        CommitHistoryDiffEntry:
        The history entries.
    """
    i = 0

    # This is not quite the same as ``enumerate(...)`` because if we run out
    # of history, ``i`` will not be incremented.

    for old_commit, new_commit in zip(old_history, new_history):
        if old_commit.commit_id != new_commit.commit_id:
            break

        yield CommitHistoryDiffEntry(
            entry_type=CommitHistoryDiffEntry.COMMIT_UNMODIFIED,
            old_commit=old_commit,
            new_commit=new_commit)

        i += 1

    for old_commit in old_history[i:]:
        yield CommitHistoryDiffEntry(
            entry_type=CommitHistoryDiffEntry.COMMIT_REMOVED,
            old_commit=old_commit)

    for new_commit in new_history[i:]:
        yield CommitHistoryDiffEntry(
            entry_type=CommitHistoryDiffEntry.COMMIT_ADDED,
            new_commit=new_commit)


def get_base_and_tip_commits(base_commit_id, tip_commit_id, diffset=None,
                             commits=None):
    """Return the base and tip commits specified.

    Args:
        base_commit_id (int):
            The primary key of the requested base commit. This may be ``None``,
            in which case a base commit will not be looked up or returned.

        tip_commit_id (int):
            The primary key of the requested tip commit. This may be ``None``,
            in which case a tip commit will not be looked up or returned.

        diffset (reviewboard.diffviewer.models.diffset.DiffSet, optional):
            The diffset that the commits belong to.

            This argument is only required if ``commits`` is ``None``.

        commits (list of reviewboard.diffviewer.models.diffcommit.DiffCommit,
                 optional):
            A pre-fetched list of commits to use instead of querying the
            database.

            If this argument is not provided, ``diffset`` must be provided to
            limit the database query to that DiffSet's commits.

    Returns:
        tuple:
        A 2-tuple of the following:

        * The requested base commit (:py:class:`~reviewboard.diffviewer.models.
          diffcommit.DiffCommit`).
        * The requested tip commit (:py:class:`~reviewboard.diffviewer.models.
          diffcommit.DiffCommit`).

        If either the base or tip commit are not requested or they cannot be
        found, then their corresponding entry in the tuple will be ``None``.
    """
    if commits is None:
        if diffset is None:
            raise ValueError(
                'get_base_and_tip_commits() requires either diffset or '
                'commits to be provided.')

        commit_ids = []

        if base_commit_id is not None:
            commit_ids.append(base_commit_id)

        if tip_commit_id is not None:
            commit_ids.append(tip_commit_id)

        if commit_ids:
            commits = list(diffset.commits.filter(pk__in=commit_ids))

    if not commits:
        return None, None

    base_commit = None
    tip_commit = None

    if base_commit_id is not None or tip_commit_id is not None:
        for commit in commits:
            if base_commit_id == commit.pk:
                base_commit = commit

            if tip_commit_id == commit.pk:
                tip_commit = commit

    return base_commit, tip_commit



from django.contrib import admin
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import DiffLexer

from reviewboard.admin import ModelAdmin, admin_site
from reviewboard.diffviewer.models import (DiffCommit, DiffSet, DiffSetHistory,
                                           FileDiff)


class _FileDiffCommon(object):
    """Common attributes for FileDiffAdmin and FileDiffInline."""

    fieldsets = (
        (None, {
            'fields': ('diffset', 'status', 'binary',
                       ('source_file', 'source_revision'),
                       ('dest_file', 'dest_detail'),
                       'diff', 'parent_diff')
        }),
        (_('Internal State'), {
            'description': _('<p>This is advanced state that should not be '
                             'modified unless something is wrong.</p>'),
            'fields': ('extra_data',),
            'classes': ['collapse'],
        }),
    )
    raw_id_fields = (
        'commit',
        'diff_hash',
        'diffset',
        'legacy_diff_hash',
        'legacy_parent_diff_hash',
        'parent_diff_hash',
    )
    readonly_fields = ('diff', 'parent_diff')

    def diff(self, filediff):
        return self._style_diff(filediff.diff)
    diff.short_description = _('Diff')

    def parent_diff(self, filediff):
        return self._style_diff(filediff.parent_diff)
    parent_diff.short_description = _('Parent diff')

    def _style_diff(self, diff):
        """Return a syntax-highlighted version of the diff.

        Args:
            diff (bytes):
                The raw diff content.

        Returns:
            django.utils.safestring.SafeText:
            The syntax-highlighted HTML.
        """
        # NOTE: Django wraps the contents in a <p>, but browsers will
        #       be sad about that, because it contains a <pre>. Chrome,
        #       for instance, will move it out into its own node. Be
        #       consistent and just make that happen for them.
        return format_html(
            '</p>{0}<p>',
            mark_safe(highlight(diff, DiffLexer(), HtmlFormatter())))


class FileDiffAdmin(_FileDiffCommon, ModelAdmin):
    """Administration UI information for FileDiff."""

    list_display = ('source_file', 'source_revision',
                    'dest_file', 'dest_detail')


class FileDiffInline(_FileDiffCommon, admin.StackedInline):
    """Inline relation information for FileDiff."""

    model = FileDiff
    extra = 0


class _DiffCommitCommon(object):
    """Common attributes for DiffCommitAdmin and DiffCommitInline."""

    raw_id_fields = ('diffset',)


class DiffCommitAdmin(_DiffCommitCommon, ModelAdmin):
    """Administration UI information for DiffCommit."""

    list_display = ('__str__',)
    inlines = (FileDiffInline,)


class DiffCommitInline(_DiffCommitCommon, admin.StackedInline):
    """Inline relation information for DiffCommit."""

    model = DiffCommit
    extra = 0
    inlines = (FileDiffInline,)


class _DiffSetCommon(object):
    """Common attributes for DiffSetAdmin and DiffSetInline."""

    raw_id_fields = ('history',)
    ordering = ('-timestamp',)


class DiffSetAdmin(_DiffSetCommon, ModelAdmin):
    """Administration UI information for DiffSet."""

    list_display = ('__str__', 'revision', 'timestamp')
    inlines = (DiffCommitInline, FileDiffInline)


class DiffSetInline(_DiffSetCommon, admin.StackedInline):
    """Inline relation information for DiffSet."""

    model = DiffSet
    extra = 0


class DiffSetHistoryAdmin(ModelAdmin):
    list_display = ('__str__', 'timestamp')
    inlines = (DiffSetInline,)
    ordering = ('-timestamp',)


admin_site.register(DiffCommit, DiffCommitAdmin)
admin_site.register(DiffSet, DiffSetAdmin)
admin_site.register(DiffSetHistory, DiffSetHistoryAdmin)
admin_site.register(FileDiff, FileDiffAdmin)



"""Argument parsing utilities for command line tools."""

import argparse
import os
import sys

import reviewboard
from reviewboard.cmdline.utils.console import get_console


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Formats help text by preserving paragraphs."""

    indent_len = 2

    def _fill_text(self, text, width, indent):
        """Return wrapped description text.

        This will wrap each contained paragraph (separated by a newline)
        individually.

        Args:
            text (unicode):
                The text to wrap.

            width (int, unused):
                The terminal width.

            indent (unicode, unused):
                The string to prefix each line with, for indentation.

        Returns:
            unicode:
            The wrapped text.
        """
        console = get_console()
        indent_len_str = ' ' * self.indent_len

        return '\n'.join(
            console.wrap_text(paragraph, indent=indent or indent_len_str)
            for paragraph in text.split('\n')
        )


class RBProgVersionAction(argparse.Action):
    """Display the Review Board/command version.

    This is used instead of :py:mod:`argparse`'s default version handling
    in order to print text unindented and unwrapped.
    """

    def __init__(self, **kwargs):
        """Initialize the action.

        Args:
            **kwargs (dict):
                Keyword arguments for the action.
        """
        super(RBProgVersionAction, self).__init__(nargs=0, **kwargs)

    def __call__(self, parser, *args, **kwargs):
        """Call the action.

        This will display the version information directly to the terminal
        and then exit.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser that called this action.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.
        """
        parser.exit(message=('\n'.join([
            'Review Board/%s %s' % (parser.prog,
                                    reviewboard.get_version_string()),
            'Python %s' % sys.version.splitlines()[0],
            'Installed to %s' % os.path.dirname(reviewboard.__file__),
            '',
        ])))



"""Django app information for reviewboard.site."""

from django.apps import AppConfig


class SiteAppConfig(AppConfig):
    """App configuration for reviewboard.site."""

    name = 'reviewboard.site'

    def ready(self):
        """Configure the app once it's ready.

        This will connect signal handlers for the app.
        """
        from reviewboard.site.signal_handlers import connect_signal_handlers

        connect_signal_handlers()
